/*
 * Copyright (C) 2012 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Aug 30, 2012
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CGAUSSQUADRATURE_2D_TRIANGLE_AREA_HPP_
#define CGAUSSQUADRATURE_2D_TRIANGLE_AREA_HPP_


namespace sierpi
{
namespace pdelab
{


/**
 * order of gaussian quadrature is equal to the number of sampling points used for the quadrature
 *
 * degree specifies the polynomial degree for which this integration accurate.
 */

template <typename T, int t_degreeOfPolynomial>
class CGaussQuadrature2D_TriangleArea
{
	static inline T* getCoords();
	static inline T* getWeights();
	static int getNumCoords();
};



/*
 *
 * We use Dunavant integration points:
 *
 * http://people.sc.fsu.edu/~jburkardt/f_src/dunavant/dunavant.html
 *
 * Also see "High degree efficient symmetrical gaussian quadrature rules for the triangle", Dunavant
 *
 * Quadrature Formulas in Two Dimensions
 * Math 5172 - Finite Element Method
 * Section 001, Spring 2010
 * author: shaodeng
 *
 * and
 *
 * "A Pragmatic Introduction to the Finite Element Method for Thermal and Stress Analysis"
 * author: (Petr Krysl)
 */

/**
 * Degree 0
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 0>	:
	public CGaussQuadrature2D_TriangleArea<T, 1>
{
};

/**
 * Degree 1
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 1>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			1
		};
		return values;
	}


	static int getNumCoords()
	{
		return 1;
	}
};

/**
 * Degree 2
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 2>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.666666666666667,	0.166666666666667,
			0.166666666666667,	0.166666666666667,
			0.166666666666667,	0.666666666666667
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.333333333333333,
			0.333333333333333,
			0.333333333333333
		};
		return values;
	}


	static int getNumCoords()
	{
		return 3;
	}
};

/**
 * Degree 3
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 3>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.6,	0.2,
			0.2,	0.2,
			0.2,	0.6
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			-0.5625,
			0.520833333333333,
			0.520833333333333,
			0.520833333333333
		};
		return values;
	}


	static int getNumCoords()
	{
		return 4;
	}
};

/**
 * Degree 4
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 4>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.10810301816807,	0.445948490915965,
			0.445948490915965,	0.445948490915965,
			0.445948490915965,	0.10810301816807,
			0.816847572980459,	0.09157621350977101,
			0.09157621350977101,	0.09157621350977101,
			0.09157621350977101,	0.816847572980459
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.223381589678011,
			0.223381589678011,
			0.223381589678011,
			0.109951743655322,
			0.109951743655322,
			0.109951743655322
		};
		return values;
	}


	static int getNumCoords()
	{
		return 6;
	}
};

/**
 * Degree 5
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 5>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.05971587178977,	0.470142064105115,
			0.470142064105115,	0.470142064105115,
			0.470142064105115,	0.05971587178977,
			0.797426985353087,	0.101286507323456,
			0.101286507323456,	0.101286507323456,
			0.101286507323456,	0.797426985353087
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.225,
			0.132394152788506,
			0.132394152788506,
			0.132394152788506,
			0.125939180544827,
			0.125939180544827,
			0.125939180544827
		};
		return values;
	}


	static int getNumCoords()
	{
		return 7;
	}
};

/**
 * Degree 6
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 6>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.501426509658179,	0.24928674517091,
			0.24928674517091,	0.24928674517091,
			0.24928674517091,	0.501426509658179,
			0.873821971016996,	0.063089014491502,
			0.063089014491502,	0.063089014491502,
			0.063089014491502,	0.873821971016996,
			0.053145049844817,	0.310352451033784,
			0.310352451033784,	0.636502499121399,
			0.636502499121399,	0.053145049844817,
			0.310352451033784,	0.053145049844817,
			0.636502499121399,	0.310352451033784,
			0.053145049844817,	0.636502499121399
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.116786275726379,
			0.116786275726379,
			0.116786275726379,
			0.050844906370207,
			0.050844906370207,
			0.050844906370207,
			0.082851075618374,
			0.082851075618374,
			0.082851075618374,
			0.082851075618374,
			0.082851075618374,
			0.082851075618374
		};
		return values;
	}


	static int getNumCoords()
	{
		return 12;
	}
};

/**
 * Degree 7
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 7>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.47930806784192,	0.26034596607904,
			0.26034596607904,	0.26034596607904,
			0.26034596607904,	0.47930806784192,
			0.869739794195568,	0.06513010290221601,
			0.06513010290221601,	0.06513010290221601,
			0.06513010290221601,	0.869739794195568,
			0.048690315425316,	0.312865496004874,
			0.312865496004874,	0.63844418856981,
			0.63844418856981,	0.048690315425316,
			0.312865496004874,	0.048690315425316,
			0.63844418856981,	0.312865496004874,
			0.048690315425316,	0.63844418856981
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			-0.149570044467682,
			0.175615257433208,
			0.175615257433208,
			0.175615257433208,
			0.053347235608838,
			0.053347235608838,
			0.053347235608838,
			0.077113760890257,
			0.077113760890257,
			0.077113760890257,
			0.077113760890257,
			0.077113760890257,
			0.077113760890257
		};
		return values;
	}


	static int getNumCoords()
	{
		return 13;
	}
};

/**
 * Degree 8
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 8>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.081414823414554,	0.459292588292723,
			0.459292588292723,	0.459292588292723,
			0.459292588292723,	0.081414823414554,
			0.65886138449648,	0.17056930775176,
			0.17056930775176,	0.17056930775176,
			0.17056930775176,	0.65886138449648,
			0.898905543365938,	0.050547228317031,
			0.050547228317031,	0.050547228317031,
			0.050547228317031,	0.898905543365938,
			0.008394777409958001,	0.263112829634638,
			0.263112829634638,	0.728492392955404,
			0.728492392955404,	0.008394777409958001,
			0.263112829634638,	0.008394777409958001,
			0.728492392955404,	0.263112829634638,
			0.008394777409958001,	0.728492392955404
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.144315607677787,
			0.09509163426728499,
			0.09509163426728499,
			0.09509163426728499,
			0.103217370534718,
			0.103217370534718,
			0.103217370534718,
			0.032458497623198,
			0.032458497623198,
			0.032458497623198,
			0.027230314174435,
			0.027230314174435,
			0.027230314174435,
			0.027230314174435,
			0.027230314174435,
			0.027230314174435
		};
		return values;
	}


	static int getNumCoords()
	{
		return 16;
	}
};

/**
 * Degree 9
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 9>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.020634961602525,	0.489682519198738,
			0.489682519198738,	0.489682519198738,
			0.489682519198738,	0.020634961602525,
			0.125820817014127,	0.437089591492937,
			0.437089591492937,	0.437089591492937,
			0.437089591492937,	0.125820817014127,
			0.623592928761935,	0.188203535619033,
			0.188203535619033,	0.188203535619033,
			0.188203535619033,	0.623592928761935,
			0.910540973211095,	0.044729513394453,
			0.044729513394453,	0.044729513394453,
			0.044729513394453,	0.910540973211095,
			0.036838412054736,	0.221962989160766,
			0.221962989160766,	0.741198598784498,
			0.741198598784498,	0.036838412054736,
			0.221962989160766,	0.036838412054736,
			0.741198598784498,	0.221962989160766,
			0.036838412054736,	0.741198598784498
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.097135796282799,
			0.031334700227139,
			0.031334700227139,
			0.031334700227139,
			0.077827541004774,
			0.077827541004774,
			0.077827541004774,
			0.07964773892721,
			0.07964773892721,
			0.07964773892721,
			0.025577675658698,
			0.025577675658698,
			0.025577675658698,
			0.043283539377289,
			0.043283539377289,
			0.043283539377289,
			0.043283539377289,
			0.043283539377289,
			0.043283539377289
		};
		return values;
	}


	static int getNumCoords()
	{
		return 19;
	}
};

/**
 * Degree 10
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 10>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.028844733232685,	0.485577633383657,
			0.485577633383657,	0.485577633383657,
			0.485577633383657,	0.028844733232685,
			0.781036849029926,	0.109481575485037,
			0.109481575485037,	0.109481575485037,
			0.109481575485037,	0.781036849029926,
			0.14170721941488,	0.307939838764121,
			0.307939838764121,	0.550352941820999,
			0.550352941820999,	0.14170721941488,
			0.307939838764121,	0.14170721941488,
			0.550352941820999,	0.307939838764121,
			0.14170721941488,	0.550352941820999,
			0.025003534762686,	0.246672560639903,
			0.246672560639903,	0.728323904597411,
			0.728323904597411,	0.025003534762686,
			0.246672560639903,	0.025003534762686,
			0.728323904597411,	0.246672560639903,
			0.025003534762686,	0.728323904597411,
			0.009540815400299,	0.0668032510122,
			0.0668032510122,	0.9236559335875,
			0.9236559335875,	0.009540815400299,
			0.0668032510122,	0.009540815400299,
			0.9236559335875,	0.0668032510122,
			0.009540815400299,	0.9236559335875
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.090817990382754,
			0.036725957756467,
			0.036725957756467,
			0.036725957756467,
			0.045321059435528,
			0.045321059435528,
			0.045321059435528,
			0.07275791684542,
			0.07275791684542,
			0.07275791684542,
			0.07275791684542,
			0.07275791684542,
			0.07275791684542,
			0.028327242531057,
			0.028327242531057,
			0.028327242531057,
			0.028327242531057,
			0.028327242531057,
			0.028327242531057,
			0.009421666963733,
			0.009421666963733,
			0.009421666963733,
			0.009421666963733,
			0.009421666963733,
			0.009421666963733
		};
		return values;
	}


	static int getNumCoords()
	{
		return 25;
	}
};

/**
 * Degree 11
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 11>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			-0.06922209654151699,	0.534611048270758,
			0.534611048270758,	0.534611048270758,
			0.534611048270758,	-0.06922209654151699,
			0.20206139406829,	0.398969302965855,
			0.398969302965855,	0.398969302965855,
			0.398969302965855,	0.20206139406829,
			0.593380199137435,	0.203309900431282,
			0.203309900431282,	0.203309900431282,
			0.203309900431282,	0.593380199137435,
			0.761298175434837,	0.119350912282581,
			0.119350912282581,	0.119350912282581,
			0.119350912282581,	0.761298175434837,
			0.935270103777448,	0.032364948111276,
			0.032364948111276,	0.032364948111276,
			0.032364948111276,	0.935270103777448,
			0.050178138310495,	0.356620648261293,
			0.356620648261293,	0.593201213428213,
			0.593201213428213,	0.050178138310495,
			0.356620648261293,	0.050178138310495,
			0.593201213428213,	0.356620648261293,
			0.050178138310495,	0.593201213428213,
			0.021022016536166,	0.171488980304042,
			0.171488980304042,	0.807489003159792,
			0.807489003159792,	0.021022016536166,
			0.171488980304042,	0.021022016536166,
			0.807489003159792,	0.171488980304042,
			0.021022016536166,	0.807489003159792
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.000927006328961,
			0.000927006328961,
			0.000927006328961,
			0.07714953491481299,
			0.07714953491481299,
			0.07714953491481299,
			0.059322977380774,
			0.059322977380774,
			0.059322977380774,
			0.036184540503418,
			0.036184540503418,
			0.036184540503418,
			0.013659731002678,
			0.013659731002678,
			0.013659731002678,
			0.052337111962204,
			0.052337111962204,
			0.052337111962204,
			0.052337111962204,
			0.052337111962204,
			0.052337111962204,
			0.020707659639141,
			0.020707659639141,
			0.020707659639141,
			0.020707659639141,
			0.020707659639141,
			0.020707659639141
		};
		return values;
	}


	static int getNumCoords()
	{
		return 27;
	}
};

/**
 * Degree 12
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 12>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.02356522045239,	0.488217389773805,
			0.488217389773805,	0.488217389773805,
			0.488217389773805,	0.02356522045239,
			0.120551215411079,	0.43972439229446,
			0.43972439229446,	0.43972439229446,
			0.43972439229446,	0.120551215411079,
			0.457579229975768,	0.271210385012116,
			0.271210385012116,	0.271210385012116,
			0.271210385012116,	0.457579229975768,
			0.7448477089168281,	0.127576145541586,
			0.127576145541586,	0.127576145541586,
			0.127576145541586,	0.7448477089168281,
			0.957365299093579,	0.02131735045321,
			0.02131735045321,	0.02131735045321,
			0.02131735045321,	0.957365299093579,
			0.115343494534698,	0.275713269685514,
			0.275713269685514,	0.608943235779788,
			0.608943235779788,	0.115343494534698,
			0.275713269685514,	0.115343494534698,
			0.608943235779788,	0.275713269685514,
			0.115343494534698,	0.608943235779788,
			0.022838332222257,	0.28132558098994,
			0.28132558098994,	0.695836086787803,
			0.695836086787803,	0.022838332222257,
			0.28132558098994,	0.022838332222257,
			0.695836086787803,	0.28132558098994,
			0.022838332222257,	0.695836086787803,
			0.02573405054833,	0.116251915907597,
			0.116251915907597,	0.858014033544073,
			0.858014033544073,	0.02573405054833,
			0.116251915907597,	0.02573405054833,
			0.858014033544073,	0.116251915907597,
			0.02573405054833,	0.858014033544073
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.025731066440455,
			0.025731066440455,
			0.025731066440455,
			0.043692544538038,
			0.043692544538038,
			0.043692544538038,
			0.06285822421788501,
			0.06285822421788501,
			0.06285822421788501,
			0.034796112930709,
			0.034796112930709,
			0.034796112930709,
			0.006166261051559,
			0.006166261051559,
			0.006166261051559,
			0.040371557766381,
			0.040371557766381,
			0.040371557766381,
			0.040371557766381,
			0.040371557766381,
			0.040371557766381,
			0.022356773202303,
			0.022356773202303,
			0.022356773202303,
			0.022356773202303,
			0.022356773202303,
			0.022356773202303,
			0.017316231108659,
			0.017316231108659,
			0.017316231108659,
			0.017316231108659,
			0.017316231108659,
			0.017316231108659
		};
		return values;
	}


	static int getNumCoords()
	{
		return 33;
	}
};

/**
 * Degree 13
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 13>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.009903630120591001,	0.495048184939705,
			0.495048184939705,	0.495048184939705,
			0.495048184939705,	0.009903630120591001,
			0.062566729780852,	0.468716635109574,
			0.468716635109574,	0.468716635109574,
			0.468716635109574,	0.062566729780852,
			0.170957326397447,	0.414521336801277,
			0.414521336801277,	0.414521336801277,
			0.414521336801277,	0.170957326397447,
			0.541200855914337,	0.229399572042831,
			0.229399572042831,	0.229399572042831,
			0.229399572042831,	0.541200855914337,
			0.77115100960734,	0.11442449519633,
			0.11442449519633,	0.11442449519633,
			0.11442449519633,	0.77115100960734,
			0.950377217273082,	0.024811391363459,
			0.024811391363459,	0.024811391363459,
			0.024811391363459,	0.950377217273082,
			0.09485382837957899,	0.268794997058761,
			0.268794997058761,	0.63635117456166,
			0.63635117456166,	0.09485382837957899,
			0.268794997058761,	0.09485382837957899,
			0.63635117456166,	0.268794997058761,
			0.09485382837957899,	0.63635117456166,
			0.018100773278807,	0.291730066734288,
			0.291730066734288,	0.690169159986905,
			0.690169159986905,	0.018100773278807,
			0.291730066734288,	0.018100773278807,
			0.690169159986905,	0.291730066734288,
			0.018100773278807,	0.690169159986905,
			0.02223307667409,	0.126357385491669,
			0.126357385491669,	0.851409537834241,
			0.851409537834241,	0.02223307667409,
			0.126357385491669,	0.02223307667409,
			0.851409537834241,	0.126357385491669,
			0.02223307667409,	0.851409537834241
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.052520923400802,
			0.01128014520933,
			0.01128014520933,
			0.01128014520933,
			0.031423518362454,
			0.031423518362454,
			0.031423518362454,
			0.047072502504194,
			0.047072502504194,
			0.047072502504194,
			0.047363586536355,
			0.047363586536355,
			0.047363586536355,
			0.031167529045794,
			0.031167529045794,
			0.031167529045794,
			0.007975771465074,
			0.007975771465074,
			0.007975771465074,
			0.036848402728732,
			0.036848402728732,
			0.036848402728732,
			0.036848402728732,
			0.036848402728732,
			0.036848402728732,
			0.017401463303822,
			0.017401463303822,
			0.017401463303822,
			0.017401463303822,
			0.017401463303822,
			0.017401463303822,
			0.015521786839045,
			0.015521786839045,
			0.015521786839045,
			0.015521786839045,
			0.015521786839045,
			0.015521786839045
		};
		return values;
	}


	static int getNumCoords()
	{
		return 37;
	}
};

/**
 * Degree 14
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 14>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.022072179275643,	0.488963910362179,
			0.488963910362179,	0.488963910362179,
			0.488963910362179,	0.022072179275643,
			0.164710561319092,	0.417644719340454,
			0.417644719340454,	0.417644719340454,
			0.417644719340454,	0.164710561319092,
			0.453044943382323,	0.273477528308839,
			0.273477528308839,	0.273477528308839,
			0.273477528308839,	0.453044943382323,
			0.645588935174913,	0.177205532412543,
			0.177205532412543,	0.177205532412543,
			0.177205532412543,	0.645588935174913,
			0.876400233818255,	0.061799883090873,
			0.061799883090873,	0.061799883090873,
			0.061799883090873,	0.876400233818255,
			0.961218077502598,	0.019390961248701,
			0.019390961248701,	0.019390961248701,
			0.019390961248701,	0.961218077502598,
			0.057124757403648,	0.172266687821356,
			0.172266687821356,	0.770608554774996,
			0.770608554774996,	0.057124757403648,
			0.172266687821356,	0.057124757403648,
			0.770608554774996,	0.172266687821356,
			0.057124757403648,	0.770608554774996,
			0.092916249356972,	0.336861459796345,
			0.336861459796345,	0.570222290846683,
			0.570222290846683,	0.092916249356972,
			0.336861459796345,	0.092916249356972,
			0.570222290846683,	0.336861459796345,
			0.092916249356972,	0.570222290846683,
			0.014646950055654,	0.298372882136258,
			0.298372882136258,	0.686980167808088,
			0.686980167808088,	0.014646950055654,
			0.298372882136258,	0.014646950055654,
			0.686980167808088,	0.298372882136258,
			0.014646950055654,	0.686980167808088,
			0.001268330932872,	0.118974497696957,
			0.118974497696957,	0.879757171370171,
			0.879757171370171,	0.001268330932872,
			0.118974497696957,	0.001268330932872,
			0.879757171370171,	0.118974497696957,
			0.001268330932872,	0.879757171370171
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.021883581369429,
			0.021883581369429,
			0.021883581369429,
			0.032788353544125,
			0.032788353544125,
			0.032788353544125,
			0.051774104507292,
			0.051774104507292,
			0.051774104507292,
			0.042162588736993,
			0.042162588736993,
			0.042162588736993,
			0.014433699669777,
			0.014433699669777,
			0.014433699669777,
			0.0049234036024,
			0.0049234036024,
			0.0049234036024,
			0.024665753212564,
			0.024665753212564,
			0.024665753212564,
			0.024665753212564,
			0.024665753212564,
			0.024665753212564,
			0.038571510787061,
			0.038571510787061,
			0.038571510787061,
			0.038571510787061,
			0.038571510787061,
			0.038571510787061,
			0.014436308113534,
			0.014436308113534,
			0.014436308113534,
			0.014436308113534,
			0.014436308113534,
			0.014436308113534,
			0.005010228838501,
			0.005010228838501,
			0.005010228838501,
			0.005010228838501,
			0.005010228838501,
			0.005010228838501
		};
		return values;
	}


	static int getNumCoords()
	{
		return 42;
	}
};

/**
 * Degree 15
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 15>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			-0.013945833716486,	0.506972916858243,
			0.506972916858243,	0.506972916858243,
			0.506972916858243,	-0.013945833716486,
			0.137187291433955,	0.431406354283023,
			0.431406354283023,	0.431406354283023,
			0.431406354283023,	0.137187291433955,
			0.444612710305711,	0.277693644847144,
			0.277693644847144,	0.277693644847144,
			0.277693644847144,	0.444612710305711,
			0.747070217917492,	0.126464891041254,
			0.126464891041254,	0.126464891041254,
			0.126464891041254,	0.747070217917492,
			0.8583832280506281,	0.070808385974686,
			0.070808385974686,	0.070808385974686,
			0.070808385974686,	0.8583832280506281,
			0.9620696595178529,	0.018965170241073,
			0.018965170241073,	0.018965170241073,
			0.018965170241073,	0.9620696595178529,
			0.133734161966621,	0.261311371140087,
			0.261311371140087,	0.604954466893291,
			0.604954466893291,	0.133734161966621,
			0.261311371140087,	0.133734161966621,
			0.604954466893291,	0.261311371140087,
			0.133734161966621,	0.604954466893291,
			0.036366677396917,	0.388046767090269,
			0.388046767090269,	0.575586555512814,
			0.575586555512814,	0.036366677396917,
			0.388046767090269,	0.036366677396917,
			0.575586555512814,	0.388046767090269,
			0.036366677396917,	0.575586555512814,
			-0.010174883126571,	0.285712220049916,
			0.285712220049916,	0.724462663076655,
			0.724462663076655,	-0.010174883126571,
			0.285712220049916,	-0.010174883126571,
			0.724462663076655,	0.285712220049916,
			-0.010174883126571,	0.724462663076655,
			0.036843869875878,	0.215599664072284,
			0.215599664072284,	0.747556466051838,
			0.747556466051838,	0.036843869875878,
			0.215599664072284,	0.036843869875878,
			0.747556466051838,	0.215599664072284,
			0.036843869875878,	0.747556466051838,
			0.012459809331199,	0.103575616576386,
			0.103575616576386,	0.883964574092416,
			0.883964574092416,	0.012459809331199,
			0.103575616576386,	0.012459809331199,
			0.883964574092416,	0.103575616576386,
			0.012459809331199,	0.883964574092416
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.001916875642849,
			0.001916875642849,
			0.001916875642849,
			0.044249027271145,
			0.044249027271145,
			0.044249027271145,
			0.051186548718852,
			0.051186548718852,
			0.051186548718852,
			0.023687735870688,
			0.023687735870688,
			0.023687735870688,
			0.013289775690021,
			0.013289775690021,
			0.013289775690021,
			0.004748916608192,
			0.004748916608192,
			0.004748916608192,
			0.038550072599593,
			0.038550072599593,
			0.038550072599593,
			0.038550072599593,
			0.038550072599593,
			0.038550072599593,
			0.027215814320624,
			0.027215814320624,
			0.027215814320624,
			0.027215814320624,
			0.027215814320624,
			0.027215814320624,
			0.002182077366797,
			0.002182077366797,
			0.002182077366797,
			0.002182077366797,
			0.002182077366797,
			0.002182077366797,
			0.021505319847731,
			0.021505319847731,
			0.021505319847731,
			0.021505319847731,
			0.021505319847731,
			0.021505319847731,
			0.007673942631049,
			0.007673942631049,
			0.007673942631049,
			0.007673942631049,
			0.007673942631049,
			0.007673942631049
		};
		return values;
	}


	static int getNumCoords()
	{
		return 48;
	}
};

/**
 * Degree 16
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 16>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.005238916103123,	0.497380541948438,
			0.497380541948438,	0.497380541948438,
			0.497380541948438,	0.005238916103123,
			0.173061122901295,	0.413469438549352,
			0.413469438549352,	0.413469438549352,
			0.413469438549352,	0.173061122901295,
			0.059082801866017,	0.470458599066991,
			0.470458599066991,	0.470458599066991,
			0.470458599066991,	0.059082801866017,
			0.518892500060958,	0.240553749969521,
			0.240553749969521,	0.240553749969521,
			0.240553749969521,	0.518892500060958,
			0.704068411554854,	0.147965794222573,
			0.147965794222573,	0.147965794222573,
			0.147965794222573,	0.704068411554854,
			0.849069624685052,	0.07546518765747399,
			0.07546518765747399,	0.07546518765747399,
			0.07546518765747399,	0.849069624685052,
			0.96680719475395,	0.016596402623025,
			0.016596402623025,	0.016596402623025,
			0.016596402623025,	0.96680719475395,
			0.103575692245252,	0.296555596579887,
			0.296555596579887,	0.599868711174861,
			0.599868711174861,	0.103575692245252,
			0.296555596579887,	0.103575692245252,
			0.599868711174861,	0.296555596579887,
			0.103575692245252,	0.599868711174861,
			0.020083411655416,	0.337723063403079,
			0.337723063403079,	0.6421935249415049,
			0.6421935249415049,	0.020083411655416,
			0.337723063403079,	0.020083411655416,
			0.6421935249415049,	0.337723063403079,
			0.020083411655416,	0.6421935249415049,
			-0.004341002614139,	0.204748281642812,
			0.204748281642812,	0.7995927209713271,
			0.7995927209713271,	-0.004341002614139,
			0.204748281642812,	-0.004341002614139,
			0.7995927209713271,	0.204748281642812,
			-0.004341002614139,	0.7995927209713271,
			0.04194178646801,	0.189358492130623,
			0.189358492130623,	0.768699721401368,
			0.768699721401368,	0.04194178646801,
			0.189358492130623,	0.04194178646801,
			0.768699721401368,	0.189358492130623,
			0.04194178646801,	0.768699721401368,
			0.014317320230681,	0.08528361568265699,
			0.08528361568265699,	0.900399064086661,
			0.900399064086661,	0.014317320230681,
			0.08528361568265699,	0.014317320230681,
			0.900399064086661,	0.08528361568265699,
			0.014317320230681,	0.900399064086661
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.046875697427642,
			0.006405878578585,
			0.006405878578585,
			0.006405878578585,
			0.041710296739387,
			0.041710296739387,
			0.041710296739387,
			0.026891484250064,
			0.026891484250064,
			0.026891484250064,
			0.04213252276165,
			0.04213252276165,
			0.04213252276165,
			0.030000266842773,
			0.030000266842773,
			0.030000266842773,
			0.014200098925024,
			0.014200098925024,
			0.014200098925024,
			0.003582462351273,
			0.003582462351273,
			0.003582462351273,
			0.032773147460627,
			0.032773147460627,
			0.032773147460627,
			0.032773147460627,
			0.032773147460627,
			0.032773147460627,
			0.015298306248441,
			0.015298306248441,
			0.015298306248441,
			0.015298306248441,
			0.015298306248441,
			0.015298306248441,
			0.002386244192839,
			0.002386244192839,
			0.002386244192839,
			0.002386244192839,
			0.002386244192839,
			0.002386244192839,
			0.019084792755899,
			0.019084792755899,
			0.019084792755899,
			0.019084792755899,
			0.019084792755899,
			0.019084792755899,
			0.006850054546542,
			0.006850054546542,
			0.006850054546542,
			0.006850054546542,
			0.006850054546542,
			0.006850054546542
		};
		return values;
	}


	static int getNumCoords()
	{
		return 52;
	}
};

/**
 * Degree 17
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 17>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.005658918886452,	0.497170540556774,
			0.497170540556774,	0.497170540556774,
			0.497170540556774,	0.005658918886452,
			0.035647354750751,	0.482176322624625,
			0.482176322624625,	0.482176322624625,
			0.482176322624625,	0.035647354750751,
			0.099520061958437,	0.450239969020782,
			0.450239969020782,	0.450239969020782,
			0.450239969020782,	0.099520061958437,
			0.199467521245206,	0.400266239377397,
			0.400266239377397,	0.400266239377397,
			0.400266239377397,	0.199467521245206,
			0.495717464058095,	0.252141267970953,
			0.252141267970953,	0.252141267970953,
			0.252141267970953,	0.495717464058095,
			0.675905990683077,	0.162047004658461,
			0.162047004658461,	0.162047004658461,
			0.162047004658461,	0.675905990683077,
			0.848248235478508,	0.075875882260746,
			0.075875882260746,	0.075875882260746,
			0.075875882260746,	0.848248235478508,
			0.968690546064356,	0.015654726967822,
			0.015654726967822,	0.015654726967822,
			0.015654726967822,	0.968690546064356,
			0.010186928826919,	0.334319867363658,
			0.334319867363658,	0.655493203809423,
			0.655493203809423,	0.010186928826919,
			0.334319867363658,	0.010186928826919,
			0.655493203809423,	0.334319867363658,
			0.010186928826919,	0.655493203809423,
			0.135440871671036,	0.292221537796944,
			0.292221537796944,	0.57233759053202,
			0.57233759053202,	0.135440871671036,
			0.292221537796944,	0.135440871671036,
			0.57233759053202,	0.292221537796944,
			0.135440871671036,	0.57233759053202,
			0.054423924290583,	0.31957488542319,
			0.31957488542319,	0.626001190286228,
			0.626001190286228,	0.054423924290583,
			0.31957488542319,	0.054423924290583,
			0.626001190286228,	0.31957488542319,
			0.054423924290583,	0.626001190286228,
			0.012868560833637,	0.190704224192292,
			0.190704224192292,	0.796427214974071,
			0.796427214974071,	0.012868560833637,
			0.190704224192292,	0.012868560833637,
			0.796427214974071,	0.190704224192292,
			0.012868560833637,	0.796427214974071,
			0.067165782413524,	0.180483211648746,
			0.180483211648746,	0.752351005937729,
			0.752351005937729,	0.067165782413524,
			0.180483211648746,	0.067165782413524,
			0.752351005937729,	0.180483211648746,
			0.067165782413524,	0.752351005937729,
			0.014663182224828,	0.080711313679564,
			0.080711313679564,	0.904625504095608,
			0.904625504095608,	0.014663182224828,
			0.080711313679564,	0.014663182224828,
			0.904625504095608,	0.080711313679564,
			0.014663182224828,	0.904625504095608
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.033437199290803,
			0.005093415440507,
			0.005093415440507,
			0.005093415440507,
			0.014670864527638,
			0.014670864527638,
			0.014670864527638,
			0.024350878353672,
			0.024350878353672,
			0.024350878353672,
			0.031107550868969,
			0.031107550868969,
			0.031107550868969,
			0.03125711121862,
			0.03125711121862,
			0.03125711121862,
			0.024815654339665,
			0.024815654339665,
			0.024815654339665,
			0.014056073070557,
			0.014056073070557,
			0.014056073070557,
			0.003194676173779,
			0.003194676173779,
			0.003194676173779,
			0.008119655318993,
			0.008119655318993,
			0.008119655318993,
			0.008119655318993,
			0.008119655318993,
			0.008119655318993,
			0.026805742283163,
			0.026805742283163,
			0.026805742283163,
			0.026805742283163,
			0.026805742283163,
			0.026805742283163,
			0.018459993210822,
			0.018459993210822,
			0.018459993210822,
			0.018459993210822,
			0.018459993210822,
			0.018459993210822,
			0.008476868534328,
			0.008476868534328,
			0.008476868534328,
			0.008476868534328,
			0.008476868534328,
			0.008476868534328,
			0.018292796770025,
			0.018292796770025,
			0.018292796770025,
			0.018292796770025,
			0.018292796770025,
			0.018292796770025,
			0.006665632004165,
			0.006665632004165,
			0.006665632004165,
			0.006665632004165,
			0.006665632004165,
			0.006665632004165
		};
		return values;
	}


	static int getNumCoords()
	{
		return 61;
	}
};

/**
 * Degree 18
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 18>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.013310382738157,	0.493344808630921,
			0.493344808630921,	0.493344808630921,
			0.493344808630921,	0.013310382738157,
			0.061578811516086,	0.469210594241957,
			0.469210594241957,	0.469210594241957,
			0.469210594241957,	0.061578811516086,
			0.127437208225989,	0.436281395887006,
			0.436281395887006,	0.436281395887006,
			0.436281395887006,	0.127437208225989,
			0.210307658653168,	0.394846170673416,
			0.394846170673416,	0.394846170673416,
			0.394846170673416,	0.210307658653168,
			0.500410862393686,	0.249794568803157,
			0.249794568803157,	0.249794568803157,
			0.249794568803157,	0.500410862393686,
			0.677135612512315,	0.161432193743843,
			0.161432193743843,	0.161432193743843,
			0.161432193743843,	0.677135612512315,
			0.846803545029257,	0.076598227485371,
			0.076598227485371,	0.076598227485371,
			0.076598227485371,	0.846803545029257,
			0.9514951212931,	0.02425243935345,
			0.02425243935345,	0.02425243935345,
			0.02425243935345,	0.9514951212931,
			0.913707265566071,	0.043146367216965,
			0.043146367216965,	0.043146367216965,
			0.043146367216965,	0.913707265566071,
			0.00843053620242,	0.358911494940944,
			0.358911494940944,	0.6326579688566361,
			0.6326579688566361,	0.00843053620242,
			0.358911494940944,	0.00843053620242,
			0.6326579688566361,	0.358911494940944,
			0.00843053620242,	0.6326579688566361,
			0.131186551737188,	0.294402476751957,
			0.294402476751957,	0.574410971510855,
			0.574410971510855,	0.131186551737188,
			0.294402476751957,	0.131186551737188,
			0.574410971510855,	0.294402476751957,
			0.131186551737188,	0.574410971510855,
			0.050203151565675,	0.325017801641814,
			0.325017801641814,	0.6247790467925119,
			0.6247790467925119,	0.050203151565675,
			0.325017801641814,	0.050203151565675,
			0.6247790467925119,	0.325017801641814,
			0.050203151565675,	0.6247790467925119,
			0.066329263810916,	0.184737559666046,
			0.184737559666046,	0.748933176523037,
			0.748933176523037,	0.066329263810916,
			0.184737559666046,	0.066329263810916,
			0.748933176523037,	0.184737559666046,
			0.066329263810916,	0.748933176523037,
			0.011996194566236,	0.218796800013321,
			0.218796800013321,	0.769207005420443,
			0.769207005420443,	0.011996194566236,
			0.218796800013321,	0.011996194566236,
			0.769207005420443,	0.218796800013321,
			0.011996194566236,	0.769207005420443,
			0.014858100590125,	0.101179597136408,
			0.101179597136408,	0.8839623022734669,
			0.8839623022734669,	0.014858100590125,
			0.101179597136408,	0.014858100590125,
			0.8839623022734669,	0.101179597136408,
			0.014858100590125,	0.8839623022734669,
			-0.035222015287949,	0.020874755282586,
			0.020874755282586,	1.014347260005363,
			1.014347260005363,	-0.035222015287949,
			0.020874755282586,	-0.035222015287949,
			1.014347260005363,	0.020874755282586,
			-0.035222015287949,	1.014347260005363
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.030809939937647,
			0.009072436679404,
			0.009072436679404,
			0.009072436679404,
			0.018761316939594,
			0.018761316939594,
			0.018761316939594,
			0.019441097985477,
			0.019441097985477,
			0.019441097985477,
			0.02775394861081,
			0.02775394861081,
			0.02775394861081,
			0.032256225351457,
			0.032256225351457,
			0.032256225351457,
			0.025074032616922,
			0.025074032616922,
			0.025074032616922,
			0.015271927971832,
			0.015271927971832,
			0.015271927971832,
			0.006793922022963,
			0.006793922022963,
			0.006793922022963,
			-0.00222309872992,
			-0.00222309872992,
			-0.00222309872992,
			0.006331914076406,
			0.006331914076406,
			0.006331914076406,
			0.006331914076406,
			0.006331914076406,
			0.006331914076406,
			0.027257538049138,
			0.027257538049138,
			0.027257538049138,
			0.027257538049138,
			0.027257538049138,
			0.027257538049138,
			0.017676785649465,
			0.017676785649465,
			0.017676785649465,
			0.017676785649465,
			0.017676785649465,
			0.017676785649465,
			0.01837948463807,
			0.01837948463807,
			0.01837948463807,
			0.01837948463807,
			0.01837948463807,
			0.01837948463807,
			0.008104732808191999,
			0.008104732808191999,
			0.008104732808191999,
			0.008104732808191999,
			0.008104732808191999,
			0.008104732808191999,
			0.007634129070725,
			0.007634129070725,
			0.007634129070725,
			0.007634129070725,
			0.007634129070725,
			0.007634129070725,
			4.6187660794e-05,
			4.6187660794e-05,
			4.6187660794e-05,
			4.6187660794e-05,
			4.6187660794e-05,
			4.6187660794e-05
		};
		return values;
	}


	static int getNumCoords()
	{
		return 70;
	}
};

/**
 * Degree 19
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 19>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			0.020780025853987,	0.489609987073006,
			0.489609987073006,	0.489609987073006,
			0.489609987073006,	0.020780025853987,
			0.090926214604215,	0.454536892697893,
			0.454536892697893,	0.454536892697893,
			0.454536892697893,	0.090926214604215,
			0.197166638701138,	0.401416680649431,
			0.401416680649431,	0.401416680649431,
			0.401416680649431,	0.197166638701138,
			0.488896691193805,	0.255551654403098,
			0.255551654403098,	0.255551654403098,
			0.255551654403098,	0.488896691193805,
			0.645844115695741,	0.17707794215213,
			0.17707794215213,	0.17707794215213,
			0.17707794215213,	0.645844115695741,
			0.779877893544096,	0.110061053227952,
			0.110061053227952,	0.110061053227952,
			0.110061053227952,	0.779877893544096,
			0.888942751496321,	0.05552862425184,
			0.05552862425184,	0.05552862425184,
			0.05552862425184,	0.888942751496321,
			0.974756272445543,	0.012621863777229,
			0.012621863777229,	0.012621863777229,
			0.012621863777229,	0.974756272445543,
			0.003611417848412,	0.395754787356943,
			0.395754787356943,	0.600633794794645,
			0.600633794794645,	0.003611417848412,
			0.395754787356943,	0.003611417848412,
			0.600633794794645,	0.395754787356943,
			0.003611417848412,	0.600633794794645,
			0.13446675453078,	0.307929983880436,
			0.307929983880436,	0.557603261588784,
			0.557603261588784,	0.13446675453078,
			0.307929983880436,	0.13446675453078,
			0.557603261588784,	0.307929983880436,
			0.13446675453078,	0.557603261588784,
			0.014446025776115,	0.26456694840652,
			0.26456694840652,	0.720987025817365,
			0.720987025817365,	0.014446025776115,
			0.26456694840652,	0.014446025776115,
			0.720987025817365,	0.26456694840652,
			0.014446025776115,	0.720987025817365,
			0.046933578838178,	0.358539352205951,
			0.358539352205951,	0.594527068955871,
			0.594527068955871,	0.046933578838178,
			0.358539352205951,	0.046933578838178,
			0.594527068955871,	0.358539352205951,
			0.046933578838178,	0.594527068955871,
			0.002861120350567,	0.157807405968595,
			0.157807405968595,	0.839331473680839,
			0.839331473680839,	0.002861120350567,
			0.157807405968595,	0.002861120350567,
			0.839331473680839,	0.157807405968595,
			0.002861120350567,	0.839331473680839,
			0.223861424097916,	0.075050596975911,
			0.075050596975911,	0.701087978926173,
			0.701087978926173,	0.223861424097916,
			0.075050596975911,	0.223861424097916,
			0.701087978926173,	0.075050596975911,
			0.223861424097916,	0.701087978926173,
			0.03464707481676,	0.142421601113383,
			0.142421601113383,	0.822931324069857,
			0.822931324069857,	0.03464707481676,
			0.142421601113383,	0.03464707481676,
			0.822931324069857,	0.142421601113383,
			0.03464707481676,	0.822931324069857,
			0.010161119296278,	0.065494628082938,
			0.065494628082938,	0.924344252620784,
			0.924344252620784,	0.010161119296278,
			0.065494628082938,	0.010161119296278,
			0.924344252620784,	0.065494628082938,
			0.010161119296278,	0.924344252620784
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.032906331388919,
			0.010330731891272,
			0.010330731891272,
			0.010330731891272,
			0.022387247263016,
			0.022387247263016,
			0.022387247263016,
			0.030266125869468,
			0.030266125869468,
			0.030266125869468,
			0.030490967802198,
			0.030490967802198,
			0.030490967802198,
			0.024159212741641,
			0.024159212741641,
			0.024159212741641,
			0.016050803586801,
			0.016050803586801,
			0.016050803586801,
			0.008084580261784,
			0.008084580261784,
			0.008084580261784,
			0.002079362027485,
			0.002079362027485,
			0.002079362027485,
			0.003884876904981,
			0.003884876904981,
			0.003884876904981,
			0.003884876904981,
			0.003884876904981,
			0.003884876904981,
			0.025574160612022,
			0.025574160612022,
			0.025574160612022,
			0.025574160612022,
			0.025574160612022,
			0.025574160612022,
			0.008880903573337999,
			0.008880903573337999,
			0.008880903573337999,
			0.008880903573337999,
			0.008880903573337999,
			0.008880903573337999,
			0.016124546761731,
			0.016124546761731,
			0.016124546761731,
			0.016124546761731,
			0.016124546761731,
			0.016124546761731,
			0.002491941817491,
			0.002491941817491,
			0.002491941817491,
			0.002491941817491,
			0.002491941817491,
			0.002491941817491,
			0.018242840118951,
			0.018242840118951,
			0.018242840118951,
			0.018242840118951,
			0.018242840118951,
			0.018242840118951,
			0.010258563736199,
			0.010258563736199,
			0.010258563736199,
			0.010258563736199,
			0.010258563736199,
			0.010258563736199,
			0.003799928855302,
			0.003799928855302,
			0.003799928855302,
			0.003799928855302,
			0.003799928855302,
			0.003799928855302
		};
		return values;
	}


	static int getNumCoords()
	{
		return 73;
	}
};

/**
 * Degree 20
 */
template <typename T>
class CGaussQuadrature2D_TriangleArea<T, 20>
{
public:
	static T* getCoords()
	{
		static T values[] = {
			0.333333333333333,	0.333333333333333,
			-0.0019009287044,	0.5009504643522,
			0.5009504643522,	0.5009504643522,
			0.5009504643522,	-0.0019009287044,
			0.023574084130543,	0.488212957934729,
			0.488212957934729,	0.488212957934729,
			0.488212957934729,	0.023574084130543,
			0.089726636099435,	0.455136681950283,
			0.455136681950283,	0.455136681950283,
			0.455136681950283,	0.089726636099435,
			0.196007481363421,	0.401996259318289,
			0.401996259318289,	0.401996259318289,
			0.401996259318289,	0.196007481363421,
			0.488214180481157,	0.255892909759421,
			0.255892909759421,	0.255892909759421,
			0.255892909759421,	0.488214180481157,
			0.647023488009788,	0.176488255995106,
			0.176488255995106,	0.176488255995106,
			0.176488255995106,	0.647023488009788,
			0.791658289326483,	0.104170855336758,
			0.104170855336758,	0.104170855336758,
			0.104170855336758,	0.791658289326483,
			0.89386207231814,	0.05306896384093,
			0.05306896384093,	0.05306896384093,
			0.05306896384093,	0.89386207231814,
			0.916762569607942,	0.041618715196029,
			0.041618715196029,	0.041618715196029,
			0.041618715196029,	0.916762569607942,
			0.976836157186356,	0.011581921406822,
			0.011581921406822,	0.011581921406822,
			0.011581921406822,	0.976836157186356,
			0.048741583664839,	0.344855770229001,
			0.344855770229001,	0.60640264610616,
			0.60640264610616,	0.048741583664839,
			0.344855770229001,	0.048741583664839,
			0.60640264610616,	0.344855770229001,
			0.048741583664839,	0.60640264610616,
			0.006314115948605,	0.377843269594854,
			0.377843269594854,	0.615842614456541,
			0.615842614456541,	0.006314115948605,
			0.377843269594854,	0.006314115948605,
			0.615842614456541,	0.377843269594854,
			0.006314115948605,	0.615842614456541,
			0.134316520547348,	0.306635479062357,
			0.306635479062357,	0.559048000390295,
			0.559048000390295,	0.134316520547348,
			0.306635479062357,	0.134316520547348,
			0.559048000390295,	0.306635479062357,
			0.134316520547348,	0.559048000390295,
			0.013973893962392,	0.249419362774742,
			0.249419362774742,	0.736606743262866,
			0.736606743262866,	0.013973893962392,
			0.249419362774742,	0.013973893962392,
			0.736606743262866,	0.249419362774742,
			0.013973893962392,	0.736606743262866,
			0.07554913290976401,	0.212775724802802,
			0.212775724802802,	0.711675142287434,
			0.711675142287434,	0.07554913290976401,
			0.212775724802802,	0.07554913290976401,
			0.711675142287434,	0.212775724802802,
			0.07554913290976401,	0.711675142287434,
			-0.008368153208227,	0.146965436053239,
			0.146965436053239,	0.861402717154987,
			0.861402717154987,	-0.008368153208227,
			0.146965436053239,	-0.008368153208227,
			0.861402717154987,	0.146965436053239,
			-0.008368153208227,	0.861402717154987,
			0.026686063258714,	0.137726978828923,
			0.137726978828923,	0.835586957912363,
			0.835586957912363,	0.026686063258714,
			0.137726978828923,	0.026686063258714,
			0.835586957912363,	0.137726978828923,
			0.026686063258714,	0.835586957912363,
			0.010547719294141,	0.059696109149007,
			0.059696109149007,	0.929756171556853,
			0.929756171556853,	0.010547719294141,
			0.059696109149007,	0.010547719294141,
			0.929756171556853,	0.059696109149007,
			0.010547719294141,	0.929756171556853
		};
		return values;
	}


public:
	static T* getWeights()
	{
		static T values[] = {
			0.033057055541624,
			0.000867019185663,
			0.000867019185663,
			0.000867019185663,
			0.011660052716448,
			0.011660052716448,
			0.011660052716448,
			0.022876936356421,
			0.022876936356421,
			0.022876936356421,
			0.030448982673938,
			0.030448982673938,
			0.030448982673938,
			0.030624891725355,
			0.030624891725355,
			0.030624891725355,
			0.0243680576768,
			0.0243680576768,
			0.0243680576768,
			0.015997432032024,
			0.015997432032024,
			0.015997432032024,
			0.007698301815602,
			0.007698301815602,
			0.007698301815602,
			-0.0006320604974879999,
			-0.0006320604974879999,
			-0.0006320604974879999,
			0.001751134301193,
			0.001751134301193,
			0.001751134301193,
			0.016465839189576,
			0.016465839189576,
			0.016465839189576,
			0.016465839189576,
			0.016465839189576,
			0.016465839189576,
			0.004839033540485,
			0.004839033540485,
			0.004839033540485,
			0.004839033540485,
			0.004839033540485,
			0.004839033540485,
			0.02580490653465,
			0.02580490653465,
			0.02580490653465,
			0.02580490653465,
			0.02580490653465,
			0.02580490653465,
			0.008471091054441,
			0.008471091054441,
			0.008471091054441,
			0.008471091054441,
			0.008471091054441,
			0.008471091054441,
			0.01835491410628,
			0.01835491410628,
			0.01835491410628,
			0.01835491410628,
			0.01835491410628,
			0.01835491410628,
			0.000704404677908,
			0.000704404677908,
			0.000704404677908,
			0.000704404677908,
			0.000704404677908,
			0.000704404677908,
			0.010112684927462,
			0.010112684927462,
			0.010112684927462,
			0.010112684927462,
			0.010112684927462,
			0.010112684927462,
			0.00357390938595,
			0.00357390938595,
			0.00357390938595,
			0.00357390938595,
			0.00357390938595,
			0.00357390938595
		};
		return values;
	}


	static int getNumCoords()
	{
		return 79;
	}
};



}
}


#endif
