/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef TRAVERSATORS_CFLUX_COMM_NORMALS_DEPTH_HPP_
#define TRAVERSATORS_CFLUX_COMM_NORMALS_DEPTH_HPP_


#include "libsierpi/stacks/CSimulationStacks.hpp"
#include "libsierpi/triangle/CTriangle_Factory.hpp"
#include "libsierpi/triangle/CTriangle_Normals.hpp"

namespace sierpi
{
namespace travs
{

/**
 * adaptive refinement
 */
template <
	class t_CKernelClass,
	typename t_CTypes
>
class CFluxComm_Normals_Depth
{
public:
	typedef t_CKernelClass CKernelClass;

	/**
	 * implementation of kernel class
	 */
	CKernelClass cKernelClass;

	typedef typename CKernelClass::TReduceValue TReduceValue;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename t_CTypes::CSimulationTypes::CCellData CCellData;
	typedef typename t_CTypes::CSimulationTypes::CEdgeData CEdgeData;
	typedef typename t_CTypes::CSimulationTypes::T T;
	typedef typename t_CTypes::CSimulationStacks CSimulationStacks;

	typedef CFluxComm_Normals_Depth<t_CKernelClass,t_CTypes> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo and fifo iterator for element data
	 */
	CStackReaderTopDown<CCellData> element_data_lifo;
	CStackReaderBottomUp<CCellData> element_data_fifo;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<CEdgeData> *edge_data_comm_left_edge_stack;
	CStack<CEdgeData> *edge_data_comm_right_edge_stack;

	/**
	 * pointer to stack used for edge communication in forward order
	 */
	CStack<CEdgeData> *edge_comm_buffer;


#define SIERPI_PARALLEL_CODE	1
#include "auto/CTraversator_FluxComm_Normals_Depth_autogenerated.hpp"
#undef SIERPI_PARALLEL_CODE
#include "../CTraversatorClassInc_SetupTwoPass.hpp"


	void action_FirstPass(
			CSimulationStacks *io_cSimulationStacks
	)
	{
		assert(io_cSimulationStacks->structure_stacks.direction == io_cSimulationStacks->element_data_stacks.direction);
		assert(io_cSimulationStacks->structure_stacks.direction == CFBStacks<char>::FORWARD);

		structure_stack_out = &io_cSimulationStacks->structure_stacks.backward;
		structure_stack_out->clear();
		assert(structure_stack_out->isEmpty());

		structure_lifo_in.setup(io_cSimulationStacks->structure_stacks.forward);

		assert(!io_cSimulationStacks->structure_stacks.forward.isEmpty());

		edge_data_comm_left_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.left;
		edge_data_comm_right_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.right;

		edge_comm_buffer = &io_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(io_cSimulationStacks->element_data_stacks.forward);

		cKernelClass.traversal_pre_hook();

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Forward)(cTriangleFactory.hypNormal, cTriangleFactory.recursionDepthFirstRecMethod);
#else
		(this->*sfcRecursiveMethod_Forward)(cTriangleFactory.recursionDepthFirstRecMethod);
#endif

		cKernelClass.op_edge_edge_stack(&(io_cSimulationStacks->edge_comm_buffer));

		assert(structure_lifo_in.isEmpty());
		assert(element_data_lifo.isEmpty());
	}


	void actionMiddlePass_computeClusterBorderCFL(
			CSimulationStacks *p_cSimulationStacks,
			TReduceValue *o_reduceValue
	)	{
		cKernelClass.updateTimestepSizeWithNetUpdates(p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.right);
		cKernelClass.updateTimestepSizeWithNetUpdates(p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.left);

		*o_reduceValue = cKernelClass.getCFL1TimestepSize();
	}


	/**
	 * second pass for parallel processing (exchange stacks)
	 */
	void action_SecondPass_Parallel(
			CSimulationStacks *p_cSimulationStacks,
			T i_timestep_size
	)
	{
		structure_lifo_in.setup(p_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		assert(!p_cSimulationStacks->structure_stacks.backward.isEmpty());


		// clear "old" edge comm stacks
		p_cSimulationStacks->edge_data_comm_edge_stacks.clear();

		edge_data_comm_left_edge_stack = &p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.right;
		edge_data_comm_right_edge_stack = &p_cSimulationStacks->edge_data_comm_exchange_edge_stacks.left;


		edge_comm_buffer = &p_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		// set the timestep size
		cKernelClass.setTimestepSize(i_timestep_size);

		// SFC traversals

#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.hypNormal, cTriangleFactory.recursionDepthFirstRecMethod);
#else
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.recursionDepthFirstRecMethod);
#endif

		// post hook
		cKernelClass.traversal_post_hook();

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());

		structure_stack_out->clear();
	}



	/**
	 * second pass for serial traversal
	 */
	void action_SecondPass_Serial(
			CSimulationStacks *io_cSimulationStacks,
			T i_timestep_size
	)
	{
		structure_lifo_in.setup(io_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(io_cSimulationStacks->element_data_stacks.forward);

		assert(!io_cSimulationStacks->structure_stacks.backward.isEmpty());

		edge_data_comm_left_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.right;
		edge_data_comm_right_edge_stack = &io_cSimulationStacks->edge_data_comm_edge_stacks.left;

		edge_comm_buffer = &io_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(io_cSimulationStacks->element_data_stacks.forward);

		// set the timestep size
		cKernelClass.setTimestepSize(i_timestep_size);

		// SFC traversals
#if CONFIG_TRAVERSATORS_WITH_NORMALS_AS_INTS
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.hypNormal, cTriangleFactory.recursionDepthFirstRecMethod);
#else
		(this->*sfcRecursiveMethod_Backward)(cTriangleFactory.recursionDepthFirstRecMethod);
#endif


		// post hook
		cKernelClass.traversal_post_hook();

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());

		structure_stack_out->clear();
	}


};

}
}

#endif /* CSTRUCTURE_ADAPTIVEREFINEMENT_HPP_ */
